/*
    tstore - an archive format for archiving tar archives
    Copyright (C) 2004  Kasper Dupont

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    with permission from the author any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "tstore.h"

uint64_t store_size;
int store_fd,hash_fd;
char *namebuf;
char *storename;
int maxused=0;

int zblock(uint8_t *buf)
{
  int i;
  for (i=0;i<512;++i)
    if (buf[i]) return 0;
  return 1;
}

uint32_t put_block(uint8_t *buf)
{
  uint8_t bucket_buf[512];
  uint8_t buf2[512];
  uint64_t this_hash=hash_sector(buf);
  uint64_t bucket_number=this_hash>>16;
  uint16_t small_hash=this_hash;
  int used,j;

  if ((this_hash==0)&&zblock(buf)) return 0;

  if (pread(hash_fd,bucket_buf,512,bucket_number*512)!=512) {
    perror(namebuf);
    exit(1);
  }

  used=bucket_buf[0];
  if (used>maxused) maxused=used;

  for (j=0;j<used;++j)
    if (GET_HASH(bucket_buf+(6+7*j))==small_hash) {
      uint64_t secnum=GET_BNUM(bucket_buf+(1+7*j));
      switch (pread(store_fd,buf2,512,secnum*512)) {
      case -1:
	perror(storename);
	exit(1);
      default:
	fprintf(stderr,"%s: Incomplete pread\n",storename);
	exit(1);
      case 512:
	break;
      }
      assert(hash_sector(buf2)==this_hash);
      if (!memcmp(buf2,buf,512)) {
	/* Found a match */
	return secnum;
      }
    }

  if (used==73) {
    fprintf(stderr,"Hash overflow\n");
    exit(1);
  }

  switch (write(store_fd,buf,512)) {
  case -1:
    perror(storename);
    exit(1);
  default:
    fprintf(stderr,"%s: Incomplete write\n",storename);
    exit(1);
  case 512:
    break;
  }

  PUT_BNUM(bucket_buf+(1+7*used),store_size);
  PUT_HASH(bucket_buf+(6+7*used),small_hash);
  bucket_buf[0]=used+1;

  if (pwrite(hash_fd,bucket_buf,512,bucket_number*512)!=512) {
    perror(namebuf);
    exit(1);
  }
  return store_size++;
}

uint8_t pointers[8][512];
int usage[16];

void put_pointer(uint64_t p, int level)
{
  //if (level==1) fprintf(stderr,">>>>>%lld\n",GET_BNUM(pointers[1]));
  PUT_BNUM(pointers[level]+5*usage[level],p);
  //fprintf(stderr,"%d %d\n",level,usage[level]);
  usage[level]++;
  if (usage[level]==102) {
    put_pointer(put_block(pointers[level]),level+1);
    //fprintf(stderr,"%d: %lld\n",level+1,GET_BNUM(pointers[level+1]));
    usage[level]=0;
  }
  //if (level==1) fprintf(stderr,"<<<<<%lld\n",GET_BNUM(pointers[1]));
}

uint64_t flush_pointers(int levels)
{
  int i;
  for (i=0;i<levels;++i)
    while(usage[i])
      put_pointer(0,i);
  return GET_BNUM(pointers[levels]);
}

int main(int argc, char ** argv)
{
  uint64_t filesize=0;
  uint32_t original_size;
  uint8_t buf[512];
  int r;
  char formbuf[42];

  char *progbuf=getenv("TSTORE");
  if (progbuf) sprintf(formbuf,"%%%dlld",strlen(progbuf));

  memset(usage,0,sizeof(usage));
  memset(pointers,0,sizeof(pointers));

  storename=argv[1];

  if (argc != 3) {
    fprintf(stderr,"Uage: %s <store> <file>\n",argv[0]);
    return 1;
  }

  if (strlen(argv[2])>494) {
    fprintf(stderr,"%s: Name too long\n",argv[2]);
    return 1;
  }

  store_fd=open(argv[1],O_RDWR|O_APPEND);
  if (store_fd==-1) {
    perror(argv[1]);
    return 1;
  }

  store_size=test_store_size(store_fd,1);
  store_error(store_size,argv[1]);

  original_size=store_size;

  namebuf=malloc(strlen(argv[1])+10);
  if (!namebuf) {
    perror("Out of memory");
    return 1;
  }

  sprintf(namebuf,"%s.idx",argv[1]);

  hash_fd=open(namebuf,O_RDWR,0666);
  if (hash_fd==-1) {
    perror(namebuf);
    return 1;
  }

  buckets=size_in_blocks(hash_fd);
  if (!buckets) {
    fprintf(stderr,"Invalid hash index (rebuild?)\n");
    return 1;
  }
  init_hash();

  do {
    r=myread(0,buf,512);
    if (r==-1) {
      perror(argv[2]);
      return 1;
    }
    if (r) {
      filesize += r;
      put_pointer(put_block(buf),0);
      if (progbuf) sprintf(progbuf,formbuf,filesize);
    }
  } while(r==512);

  if (filesize) {
    uint32_t block_directory=flush_pointers(filelevels(filesize));

    --original_size;
    memset(buf,0,512);
    PUT_BNUM(buf,original_size);
    PUT_BNUM(buf+5,block_directory);
    PUT_SIZE(buf+10,filesize);
    strcpy(buf+17,argv[2]);

    while(1) {
      uint32_t direntry=put_block(buf);
      if(direntry+1 == store_size) break;
      //printf("%d,%d\n",prevsize,direntry);
      memset(buf,0,512);
      PUT_BNUM(buf,direntry);
    }
  }

  hash_advice(maxused);

  return 0;
}
